package com.vanheusden.pfa;

import java.io.*;

class AsyncBrain implements Runnable {
	Thread t;
	IO io;
	Brain brain;
	int moveNr;
	Scene scene;
	int endDepth;
	PlayerColor myColor;
	MoveRepetition history;
	long maxDuration;
	int maxExtension;
	Move lastMove;
	MoveResult sr = null;

	public void run() {
		try {
			sr = brain.calculateMove(moveNr, scene, 0/* FIXME endDepth*/, myColor, history, maxDuration, maxExtension, lastMove, false);

			io.progressCallback("AsyncBrain result: " + sr);
		}
		catch(Exception e) {
			Statics.logException("AsyncBrain", e);

			try {
				io.progressCallback("AsyncBrain thread exception: " + e);
			}
			catch(Exception e2) {
				Statics.log("Reporting an exception gave an exception: " + e2);
			}
		}
	}

	public AsyncBrain(final IO io, final Brain brain) {
		this.io = io;
		this.brain = brain;
	}

	public void startAsyncBrain(final int moveNr, final Scene scene, final int endDepth, final PlayerColor myColor, final MoveRepetition history, final long maxDuration, final int maxExtension, final Move lastMove) throws IOException {
		this.moveNr = moveNr;
		this.scene = scene;
		this.endDepth = endDepth;
		this.myColor = myColor;
		this.history = history;
		this.maxDuration = maxDuration;
		this.maxExtension = maxExtension;
		this.lastMove = lastMove;

		t = new Thread(this);
		t.start();

		io.progressCallback("AsyncBrain thread started");
	}

	public void abortSearch() throws InterruptedException {
		brain.asyncStopSearch();

		t.join();
		t = null;
	}

	// returns true when ended
	// duration in ms
	public boolean tryStopSearch(long duration) throws IOException {
		if (t == null)
			return true;

		try {
			t.join(duration);
		}
		catch(Exception e) {
			io.progressCallback("While waiting for ponder thread to stop: " + e);
		}

		return !t.isAlive();
	}

	public MoveResult getResult() {
		MoveResult temp = sr;
		sr = null;

		return temp;
	}
}
