package com.vanheusden.pfa;

import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.management.*;
import java.lang.StackTraceElement;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.imageio.ImageIO;

public class BPMWebServer implements Runnable {
	String adapter;
	int port;
	Thread t = null;
	BrainPuppetMaster bpm;
	MyHTTPServer socket = null;

	public BPMWebServer(final String adapter, final int port, final BrainPuppetMaster bpm) {
		this.adapter = adapter;
		this.port = port;
		this.bpm = bpm;

		t = new Thread(this);
		t.start();
	}

	void destructor() {
		if (t != null)
			t.interrupt();

		try {
			if (socket != null)
				socket.destructor();
		}
		catch(IOException ie) {
			Statics.log("HTTPServer: i/o exception in destructor " + ie);
		}
	}

	public void sendReply_data_json(MyHTTPServer socket) throws IOException {
		List<String> reply = new ArrayList<String>();

		reply.add("HTTP/1.0 200 OK\r\n");
		reply.add("Connection: close\r\n");
		reply.add("Content-Type: text/plain\r\n");
		reply.add("\r\n");

		int [] dronesAvail = bpm.getNumberReady();

		reply.add("{ \"nps\" : \"" + bpm.getNps() + "\", \"nnodes\" : \"" + dronesAvail[1] + "\", \"nready\" : \"" + dronesAvail[0] + "\" }");

		socket.sendReply(reply);
	}

	public void sendReply_index(MyHTTPServer socket) throws IOException {
		List<String> reply = new ArrayList<String>();

		reply.add("HTTP/1.0 200 OK\r\n");
		reply.add("Connection: close\r\n");
		reply.add("Content-Type: text/html\r\n");
		reply.add("\r\n");

		int [] dronesAvail = bpm.getNumberReady();

		reply.add("<H1>DeepBrutePos/PuppetMaster " + Version.getVersion() + "/" + Version.getRevision());
		reply.add("<TABLE BORDER=1>\n");
		reply.add("<TR><TD>nps:</TD><TD>" + bpm.peekNps() + "</TD></TR>");
		reply.add("<TR><TD># nodes:</TD><TD>" + dronesAvail[1] + "</TD></TR>");
		reply.add("<TR><TD># ready:</TD><TD>" + dronesAvail[0] + "</TD></TR>");
		reply.add("</TABLE>\n");

		reply.add("<TABLE BORDER=1>\n");
		reply.add("<TR><TH>name</TH><TH>ready</TH><TH>thread running</TH><TH>last nps</TH><TH>last duration</TH><TH>reconnects</TH><TH>time outs</TH><TH>chosen</TH><TH>not chosen</TH><TH>illegal move</TH><TH>had no idea</TH><TH>% chosen</TH></TR>\n");
		for(Drone cur : bpm.getDrones()) {
			reply.add("<TR>");
			reply.add("<TD>" + cur.getName() + "</TD>");
			reply.add("<TD>" + cur.isReady() + "</TD>");
			reply.add("<TD>" + cur.isThreadRunning() + "</TD>");
			reply.add("<TD>" + cur.peekLastNps() + "</TD>");
			reply.add("<TD>" + cur.peekThinkDuration() + "</TD>");
			reply.add("<TD>" + cur.getNReconnects() + "</TD>");
			reply.add("<TD>" + cur.getNTimeOut() + "</TD>");
			reply.add("<TD>" + cur.getNChosen() + "</TD>");
			reply.add("<TD>" + cur.getNNotChosen() + "</TD>");
			reply.add("<TD>" + cur.getNIllegalMove() + "</TD>");
			reply.add("<TD>" + cur.getNoIdea() + "</TD>");
			int n = cur.getNChosen() + cur.getNNotChosen();
			double perc = cur.getNChosen() * 100.0 / n;
			reply.add("<TD>" + String.format("%.2f%%", perc) + "</TD>");
			reply.add("</TR>\n");
		}
		reply.add("</TABLE>\n");

		socket.sendReply(reply);
	}

	public void sendReply_404(MyHTTPServer socket, String url) throws IOException {
		List<String> reply = new ArrayList<String>();

		reply.add("HTTP/1.0 404 Url not known\r\n");
		reply.add("Connection: close\r\n");
		reply.add("Content-Type: text/html\r\n");
		reply.add("\r\n");

		reply.add("URL \"" + url + "\" not known!");

		socket.sendReply(reply);
	}

	public void run() {
	stop_thread:
		for(;!Thread.interrupted();) {
			try {
				socket = new MyHTTPServer(adapter, port);
			}
			catch(IOException e) {
				Statics.log("HTTPServer Cannot create listen socket: " + e);

				try {
					Thread.sleep(1000);
				}
				catch(InterruptedException ie) {
					break stop_thread;
				}

				continue;
			}

			for(;;) {
				try {
					List<String> request = socket.acceptConnectionGetRequest();
					if (request == null) {
						Statics.log("HTTPServer: server socket closed");
						break stop_thread;
					}
					if (request.size() == 0) {
						Statics.log("HTTPServer: empty request");
						continue;
					}
					String url = request.get(0).substring(4).trim(), pars = null;
					int space = url.indexOf(" ");
					if (space != -1) {
						url = url.substring(0, space);
					}

					int q = url.indexOf("?");
					if (q != -1) {
						pars = url.substring(q + 1);
						url = url.substring(0, q);
					}

					if (url.equals("/data.json"))
						sendReply_data_json(socket);
					else if (url.equals("/") || url.equals("/index.html"))
						sendReply_index(socket);
					else
						sendReply_404(socket, url);
				}
				catch(IOException ie) {
					Statics.log("HTTPServer ioexception: " + ie);

					try {
						socket.close();
					}
					catch(Exception e2) {
						Statics.log("HTTPServer exception during server socket close: " + e2);
						Statics.logException("BPMWebServer", e2);
					}
				}
                                catch(InterruptedException ie) {
                                        break stop_thread;
                                }
				catch(Exception e) {
					Statics.log("HTTPServer exception: " + e);
					Statics.logException("BPMWebServer", e);
				}
			}
		}

		Statics.log("HTTPServer thread terminating");
	}
}
