package com.vanheusden.pfa;

import java.io.*;
import java.net.*;
import java.util.*;

class BrainPuppetMaster extends Brain {
	List<Drone> drones = null;
	int maxNCores = 0;
	static final long stopWait = 5;
	static final long droneThinkTimeAdjustment = 50 + stopWait; // FIXME 250 should depend on latency / +3: for `stop'
	static final long droneResultProcessingTimeAdjustment = 50 + stopWait;
	BPMWebServer bws = new BPMWebServer("0.0.0.0", 55000, this);
	final Object notifyObject = new Object();
	final boolean ponder;
	String brainDump = null;
	boolean firstLine = true;

	void setIO(final IO io) {
		this.io = io;

		for(Drone cur : drones)
			cur.setIO(io);
	}

	long getNps() {
		long total = 0;

		for(Drone current : drones) {
			long cn = current.getLastNps();

			if (cn != -1)
				total += cn;
		}

		return total;
	}

	long peekNps() {
		long total = 0;

		for(Drone current : drones) {
			long cn = current.peekLastNps();

			if (cn != -1)
				total += cn;
		}

		return total;
	}

	int [] getNumberReady() {
		int total = 0;

		for(Drone current : drones) {
			if (current.isReady())
				total++;
		}

		return new int[] { total, drones.size() };
	}

	List<Drone> getDrones() {
		return drones;
	}

	void emitBrainDumpInit() {
		if (firstLine) {
			firstLine = false;

			emitToBrainDumpFile("MoveNr \t" + "FEN \t" + "BOOK \t" + "eval/depth/move/pv[selected](hits) [...list of choices...]");

			emitToBrainDumpFile("MoveNr \t" + "FEN \t" + "CALC \t" + "eval/depth/move [...] chosen:e/d/m");

			String dl = "Drones:";
			for(Drone d : drones)
				dl += "\t" + d.getName();
			emitToBrainDumpFile(dl);
		}
	}

	BrainPuppetMaster(final List<Book> books, final List<Drone> drones, final int maxNCores, final boolean ponder, final String brainDump) throws Exception {
		super(null, 0, 1, 1, books, false);

		this.drones = drones;

		this.ponder = ponder;

		this.brainDump = brainDump;

		io = new IO_logonly();

		io.progressCallback("Got " + drones.size() + " drones");

		for(Drone current : drones) {
			current.setNCores(maxNCores);

			current.startThread(notifyObject);
		}

		io.progressCallback("Drones started");

		r.setSeed(System.currentTimeMillis());
	}

	void destructor() throws IOException {
		super.destructor();

		if (bws != null)
			bws.destructor();

		for(Drone current : drones)
			current.destructor();
	}

	boolean checkIfIsARepetition(final MoveRepetition history, final Scene s, final PlayerColor c, final Move m) {
		final Scene newScene;

		if (m.getScene() != null)
			newScene = m.getScene();
		else {
			newScene = s.Move(m);
			m.setScene(newScene);
		}

		final PlayerColor newColor = Statics.opponentColor(c);

		final long boardHash = getHash(newScene.getBoard(), newColor);

		return history.contains(newColor, boardHash);
	}

	void startPonder(final Scene sc, final Move lastMove, final PlayerColor c) {
		Statics.log("PuppetMaster: startPonder");
		final Scene newScene;

		if (lastMove.getScene() != null)
			newScene = lastMove.getScene();
		else {
			newScene = sc.Move(lastMove);

			lastMove.setScene(newScene);
		}

		final PlayerColor nc = Statics.opponentColor(c);

		for(Drone current : drones)
			current.startPonder(newScene, nc, lastMove);
	}

	void stopPonder() {
		Statics.log("PuppetMaster: stopPonder");
		for(Drone current : drones)
			current.stopPonder();
	}

	void asyncStopSearch() {
		Statics.log("PuppetMaster: asyncStopSearch");
		for(Drone current : drones)
			current.stopAsk(); // FIXME interruptThinking() ?
        }

	void emitToBrainDumpFile(final String what) {
		if (brainDump == null)
			return;

		Writer out = null;
		try {
			out = new OutputStreamWriter(new FileOutputStream(brainDump, true), "ascii");

			out.write(what + "\n");
		}
		catch(Exception e) {
			Statics.logException("emitToBrainDumpFile", e);
		}
		finally {
			if (out != null) {
				try {
					out.close();
				}
				catch(Exception e) {
					Statics.logException("emitToBrainDumpFile:close", e);
				}
			}
		}
	}

	void emitBrainDumpBook(final int moveNr, final Scene scene, final PlayerColor color, final Move lastMove, final List<MoveResult> allOb, final MoveResult obSel, final Long obSelHits) {
		emitBrainDumpInit();

		String out = "" + moveNr + ". \t" +
				Statics.toFENString(scene.getBoard(), color, lastMove, moveNr, true) + " " +
				"BOOK " + "\t" + obSel.toString() + "(" + obSelHits + ")";

		for(MoveResult cur : allOb)
			out += "\t" + cur.toString();

		emitToBrainDumpFile(out);
	}

	void emitBrainDumpMoves(final int moveNr, final Scene scene, final PlayerColor color, final Move lastMove, final List<MoveResult> mrs, final MoveResult chosenMove) {
		emitBrainDumpInit();

		String line = "" + moveNr + ". \t" +
			Statics.toFENString(scene.getBoard(), color, lastMove, moveNr, true) + " " +
			"CALC " + "\t";

		int bestScore = -1, bestDepth = -1;
		for(MoveResult mr : mrs) {
			String mrStr = mr.toString();

			line += mrStr;

			if (mrStr.equals(chosenMove)) {
				if (mr.getDepth() > bestDepth) {
					bestDepth = mr.getDepth();
					bestScore = mr.getEval();
				}
			}
		}

		line += chosenMove + "/" + bestScore + "/" + bestDepth;

		emitToBrainDumpFile(line);
	}

	public MoveResult calculateMove(final int moveNr, final Scene scene, final int endDepth, final PlayerColor myColor, final MoveRepetition history, final long maxDuration, final int maxExtension, final Move lastMove, final boolean isPonder) throws IOException, InterruptedException {
		long startThink = System.currentTimeMillis();
		long endThinkTime = startThink + maxDuration;

		io.progressCallback("" + moveNr + ". Puppet master: move " + moveNr);

		scene.createMoveList(myColor);
		scene.validateMoves(myColor);
		List<Move> moves = scene.getMoveList(myColor);

		io.progressCallback("" + moveNr + ". Checking book");

		List<MoveResult> mrs = new ArrayList<MoveResult>();

		boolean bookResults = false;
		if (!isPonder) {
			List<MoveResult> obResults = getOpeningBook(scene, lastMove, myColor);

			if (obResults != null) {
				Object [] obResultTemp = selectOpening(obResults);
				long nHits = obResultTemp != null ? (Long)obResultTemp[1] : 0;

				if (nHits > 1) {
					MoveResult obResult = (MoveResult)obResultTemp[0];

					io.progressCallback("Book: " + obResult + " out of " + obResults.size() + " results, with " + nHits + " hit(s)");
					io.progressCallbackXboardShowThinking(obResult.getDepth(), obResult.getEval(), (System.currentTimeMillis() - startThink) / 10, 0, obResult.getMove() + " " + obResult.getPv());

					emitBrainDumpBook(moveNr, scene, myColor, lastMove, obResults, obResult, nHits);

					return obResult;
				}

				Statics.log("Inconclusive book results, adding " + obResults.size() + " move(s) to decision list");
				mrs.addAll(obResults);
				bookResults = true;
			}
		}

		stopPonder();

		io.progressCallback("" + moveNr + ". pondering stopped (where neccessary)");

		io.progressCallback("" + moveNr + ". global data initialized");

		int invalidCnt = 0;
		int nDisconnects = 0;

		int nSuggestionsMoveRepetition = 0;

		MoveResult chosenMove = null;
		Long chosenCount = null;
		int chosenScore = 0, chosenDepth = 0;
		int nSuggestions = 0;

		int nRunning = drones.size();
		Move [] resultMoves = new Move[nRunning];
		boolean [] hasMove = new boolean[nRunning];
		PMResult [] res = new PMResult[nRunning];

		// substract some time needed for network latency and/or other
		// overhead talking to an external chess program
		long droneDeadLine = endThinkTime - droneThinkTimeAdjustment;

		boolean chosenByChosenCount = false;

		if (maxDuration >= droneThinkTimeAdjustment) {
			// start search
			for(Drone current : drones) {
				io.progressCallback("" + moveNr + ". Asking " + current);

				current.startAsk(scene, myColor, lastMove, 1, droneDeadLine);
			}

			// substract a few ms needed to do local processing
			long sleepUntil = droneDeadLine - droneResultProcessingTimeAdjustment;

			long sleepTime = Statics.getDelayUntil(sleepUntil);
			if (sleepTime < 1)
				sleepTime = 1;

			io.progressCallback("" + moveNr + ". sleeping for " + sleepTime + "ms (and waiting for any results)");

			boolean interrupted = false;

			for(;;) {
				long timeLeft = Statics.getDelayUntil(sleepUntil);

				final int droneCount = drones.size();

				if (timeLeft <= stopWait && !interrupted) {
					boolean hit = false;

					interrupted = true;

					for(int droneIndex=0; droneIndex<droneCount; droneIndex++) {
						if (!hasMove[droneIndex]) {
							Drone cur = drones.get(droneIndex);
							cur.interruptThinking();
							io.progressCallback(cur.getName() + " interrupted thinking");
							hit = true;
						}
					}

					if (hit && timeLeft > 0) {
						Thread.sleep(timeLeft);
						io.progressCallback("Asked remaining engines to stop");
					}
				}
				else if (timeLeft <= 0) {
					io.progressCallback("Out of time");
					break;
				}

				for(int droneIndex=0; droneIndex<droneCount; droneIndex++) {
					if (hasMove[droneIndex])
						continue;

					Drone currentDrone = drones.get(droneIndex);

					MoveResult move = currentDrone.getResult(1);
					if (move != null) {
						io.progressCallback("" + moveNr + ". " + currentDrone + " has decided " + move + " after " + currentDrone.getThinkDuration() + " out of " + (maxDuration - droneThinkTimeAdjustment) + " ms");

						if (move.getMove().isA1A1()) {
							hasMove[droneIndex] = true;
							res[droneIndex] = PMResult.noIdea;
							io.progressCallback("" + moveNr + ". " + currentDrone + " told us(!) it could not decide on a move");
						}
						else if (findMove(moves, move.getMove()) == -1) {
							res[droneIndex] = PMResult.illegalMove;
							io.progressCallback("" + moveNr + ". " + currentDrone + " returned an invalid move: " + move);
							invalidCnt++;
						}
						else {
							hasMove[droneIndex] = true;
							resultMoves[droneIndex] = move.getMove();
							mrs.add(move);
						}

						--nRunning;
						io.progressCallback("" + moveNr + ". Number of drones left: " + nRunning);
					}
				}

				if (nRunning == 0)
					break;

				long timeLeftAfterProbes = Statics.getDelayUntil(sleepUntil);

				if (timeLeftAfterProbes >= stopWait)
					timeLeftAfterProbes -= stopWait;

				if (timeLeftAfterProbes > 0) {
					synchronized(notifyObject) {
						notifyObject.wait(timeLeftAfterProbes);
					}
				}
			}

			io.progressCallback("" + moveNr + ". Get & process results");

			for(int droneIndex=0; droneIndex<drones.size(); droneIndex++) {
				Drone current = drones.get(droneIndex);

				if (hasMove[droneIndex])
					continue;

				MoveResult move = current.getResult(1);
				if (move == null) {
					res[droneIndex] = PMResult.timeOut;
					continue;
				}

				if (move.getMove().isA1A1()) {
					hasMove[droneIndex] = true;
					res[droneIndex] = PMResult.noIdea;
					io.progressCallback("" + moveNr + ". " + current + " told us(!) it could not decide on a move");
				}
				else if (findMove(moves, move.getMove()) == -1) {
					res[droneIndex] = PMResult.illegalMove;
					invalidCnt++;
					io.progressCallback("" + moveNr + ". " + current + " returned an invalid move " + move);
					continue;
				}

				hasMove[droneIndex] = true;

				io.progressCallback("" + moveNr + ". " + current + " chose " + move);

				resultMoves[droneIndex] = move.getMove();
				mrs.add(move);
			}

			if (nRunning > 0) {
				io.progressCallback("" + moveNr + ". Stop ask");

				// ask drones to finish
				for(int droneIndex=0; droneIndex<drones.size(); droneIndex++) {
					Drone cur = drones.get(droneIndex);

					// don't ask engines to stop that already have answered
					if (!hasMove[droneIndex]) {
						cur.stopAsk();
						res[droneIndex] = PMResult.timeOut;
						io.progressCallback(cur.getName() + " did not respond in time");

						// terminate any sessions that did not respond to
						// prevent confusion
						io.progressCallback("Disconnect " + cur);
						cur.disconnect();
						nDisconnects++;
					}
					else {
						io.progressCallback(cur.getName() + " responded");
					}
				}
			}

			// see if suggestions would cause a repetition; remove those
			for(int i=0; i<mrs.size();) {
				if (checkIfIsARepetition(history, scene, myColor, mrs.get(i).getMove())) {
					mrs.remove(i);
					nSuggestionsMoveRepetition++;
				}
				else {
					i++;
				}
			}

			// see how often a move was chosen
			Map<String, Object []> resultMap = new HashMap<String, Object []>();

			for(MoveResult rs : mrs) {
				String moveStr = rs.getMove().toString();

				Object [] value = resultMap.get(moveStr);

				if (value == null)
					value = new Object[] { new Long(1), rs };
				else {
					Long count = (Long)value[0];
					count++;
					value[0] = count;
				}

				resultMap.put(moveStr, value);
			}

			// find the move chosen most
			for(String move : resultMap.keySet()) {
				Object [] value = resultMap.get(move);
				Long count = (Long)value[0];
				MoveResult m = (MoveResult)value[1];

				if (chosenCount == null || count > chosenCount) {
					chosenCount = count;
					chosenMove = m;
				}
				else if (count == chosenCount && m.getDepth() > chosenMove.getDepth()) {
					chosenMove = m;
					chosenByChosenCount = true;
				}
				// FIXME overdo it and also check eval?

				nSuggestions++;
			}
		}

		boolean rndChosenThem = false, rndChosenGlobal = false;

		if (chosenCount == null) {
			io.progressCallback("" + moveNr + ". No moves returned by any puppet!");

			int nr = r.nextInt(moves.size());
			chosenMove = new MoveResult(-1, -1, moves.get(nr), "panic");
			chosenCount = new Long(0);

			rndChosenGlobal = true;
		}
		// if selected by one vote, chose one random
		else if (chosenCount <= 1) {
			chosenMove = mrs.get(r.nextInt(mrs.size()));

			rndChosenThem = true;
		}

		int nDidReturnSomething = 0;
		for(int droneIndex=0; droneIndex<drones.size(); droneIndex++) {
			if (res[droneIndex] == null) {
				nDidReturnSomething++;

				if (chosenMove.getMove().equals(resultMoves[droneIndex]))
					res[droneIndex] = PMResult.chosen;
				else
					res[droneIndex] = PMResult.notChosen;
			}

			drones.get(droneIndex).countResult(res[droneIndex]);
		}

		long endThink = System.currentTimeMillis();

		io.progressCallback("" + moveNr + ". Chosen " + chosenMove + ", hits: " + chosenCount + " from " + nSuggestions + " unique suggestions by " + nDidReturnSomething + " drones, random from choices: " + rndChosenThem + ", invalid: " + invalidCnt + ", disconnects: " + nDisconnects + ", random global: " + rndChosenGlobal + ", time left: " + (maxDuration - (endThink - startThink)) + "ms, moverep: " + nSuggestionsMoveRepetition + ", cbcc: " + chosenByChosenCount + ", frombook: " + bookResults);

		if (ponder) {
			io.progressCallback("Start pondering");
			startPonder(scene, chosenMove.getMove(), myColor);
		}

		io.beep();

		emitBrainDumpMoves(moveNr, scene, myColor, lastMove, mrs, chosenMove);

		return chosenMove;
	}
}
