package com.vanheusden.pfa;

import java.util.*;

class ChessObjectPawn extends ChessObject {
	public boolean equals(Object o) {
		final ChessObject oco = (ChessObject)o;

		return oco instanceof ChessObjectPawn && oco.getX() == x && oco.getY() == y && oco.getColor() == color && oco.getNMoves() == nMoves;
	}

	private void addMove(final ChessObject targetField, final int fx, final int fy, final int tx, final int ty, final ChessObjectType promoteTo) {
		if (targetField != null)
			targetField.addCatcher();

		moves.add(new Move(fx, fy, tx, ty, promoteTo));

		canDoNMoves++;
	}

	private void addPawnPromotionMove(final ChessObject targetField, final int fx, final int fy, final int tx, final int ty) {
		if (ty == 7 || ty == 0) {
			addMove(targetField, fx, fy, tx, ty, ChessObjectType.QUEEN);
			addMove(targetField, fx, fy, tx, ty, ChessObjectType.BISHOP);
			addMove(targetField, fx, fy, tx, ty, ChessObjectType.ROOK);
			addMove(targetField, fx, fy, tx, ty, ChessObjectType.KNIGHT);
		}
		else {
			addMove(targetField, fx, fy, tx, ty);
		}
	}

	public ChessObjectPawn(final int x, final int y, final PlayerColor color) {
		super(x, y, color, 0);
	}

	public ChessObjectPawn(final int x, final int y, final PlayerColor color, final int nMoves) {
		super(x, y, color, nMoves);
	}

	public ChessObject duplicate() {
		return new ChessObjectPawn(x, y, color, nMoves);
	}

	public ChessObjectType getType() {
		return ChessObjectType.PAWN;
	}

	public String getTypeCharacter(final boolean uppercase) {
		final String character = ChessObjectType.PAWN.getName();

		if (uppercase)
			return character.toUpperCase();

		return character;
	}

	public void getMoves(final Scene scene, final List<Move> moves, final ChessObject king, final Move lastMove) {
		this.moves = moves;
		this.board = scene.getBoard();

		/* en passant */
		if (lastMove != null) {
			/* is that last moved object a pawn which did it's first move? */
			/* initial move 2 in one? destination is empty? */
			if (lastMove.getTY() == y && Math.abs(lastMove.getTX() - x) == 1) {
				final int destX = lastMove.getTX();
				final int destY = color == PlayerColor.WHITE ? 5 : 2;

				ChessObject targetField = board.getAt(destX, destY);
				addMove(targetField, x, y, destX, destY, MoveType.EN_PASSANT);

				board.getAt(lastMove.getTX(), lastMove.getTY()).addCatcher();
			}
		}

		final int promotionLine;
		final int yPlus1, yPlus2;

		if (color == PlayerColor.WHITE) {
			yPlus1 = y + 1;
			yPlus2 = y + 2;
			promotionLine = 7;
		}
		else {
			yPlus1 = y - 1;
			yPlus2 = y - 2;
			promotionLine = 0;
		}

		if (y == promotionLine) {
			/* object can be promoted but it was chosen not to */
		}
		else {
			if (board.getAt(x, yPlus1) == null) {
				addPawnPromotionMove(null, x, y, x, yPlus1);

				if (isFirstMove() && board.getAt(x, yPlus2) == null)
					addMove(x, y, x, yPlus2);
			}

			final int leftX = x - 1;
			if (leftX >= 0) {
				final ChessObject obj = board.getAt(leftX, yPlus1);

				if  (obj != null && obj.getColor() != color)
					addPawnPromotionMove(obj, x, y, leftX, yPlus1);
			}

			final int rightX = x + 1;
			if (rightX <= 7) {
				final ChessObject obj = board.getAt(rightX, yPlus1);

				if (obj != null && obj.getColor() != color)
					addPawnPromotionMove(obj, x, y, rightX, yPlus1);
			}
		}

		assert canDoNMoves <= 13;

		generatedListOfMoves = true;
	}

	public ChessObject promote(final ChessObjectType newType) {
		switch(newType) {
			case PAWN:
				return new ChessObjectPawn(x, y, color, nMoves);
			case ROOK:
				return new ChessObjectRook(x, y, color, nMoves);
			case BISHOP:
				return new ChessObjectBishop(x, y, color, nMoves);
			case KNIGHT:
				return new ChessObjectKnight(x, y, color, nMoves);
			case KING:
				return new ChessObjectKing(x, y, color, nMoves);
			case QUEEN:
				return new ChessObjectQueen(x, y, color, nMoves);
		}

		assert false;

		return null;
	}
}
