package com.vanheusden.pfa;

import java.io.*;
import java.util.*;

class IO_Console extends IO {
	enum ReturnType { RET_MOVE, RET_UNDO, RET_QUIT };

	class ReturnData {
		ReturnType rt;
		Move m;

		ReturnData(final ReturnType rt, final Move m) {
			this.rt = rt;
			this.m = m;
		}

		ReturnType getReturnType() {
			return rt;
		}

		Move getMove() {
			return m;
		}
	}

	class HistoryElement {
		String fenWhiteBefore = null, fenBlackBefore = null;
		Move moveWhite = null, moveBlack = null;

		HistoryElement() {
		}

		void setWhite(String fen, Move move) {
			fenWhiteBefore = fen;
			moveWhite = move;
		}

		Move getWhiteMove() {
			return moveWhite;
		}

		String getWhiteBeforeFen() {
			return fenWhiteBefore;
		}

		void setBlack(String fen, Move move) {
			fenBlackBefore = fen;
			moveBlack = move;
		}

		Move getBlackMove() {
			return moveBlack;
		}

		String getBlackBeforeFen() {
			return fenBlackBefore;
		}
	}

	List<HistoryElement> h = new ArrayList<HistoryElement>();

	public void progressCallback(String what) {
		Statics.log(what);

		if (Statics.verbose > 0)
			System.out.println(what);
	}

	public void progressCallbackDebug(String what) {
		if (Statics.verbose > 0) {
			Statics.log(what);
			System.out.println(what);
		}
	}

	public void progressCallbackXboardShowThinking(int id, int score, long durationCentiSeconds, long nodesVisited, String details) throws IOException {
		String what = "" + id + " " + score + " " + durationCentiSeconds + " " + nodesVisited + " " + details;

		Statics.log(what);

		System.out.println(what);
	}

	public void progressCallbackNPS(int nps) throws IOException {
		String what = "Nodes/s " + nps;

		Statics.log(what);

		if (Statics.verbose > 0)
			System.out.println(what);
	}

	public void progressCallbackTimeSearch(int ms) throws IOException {
		String what = "Time searched: " + ms + "ms";

		Statics.log(what);

		if (Statics.verbose > 0)
			System.out.println(what);
	}

	public void progressCallbackBest(Move move, int score) throws IOException {
		String what = "Best move: " + move + " (" + score + ")";

		Statics.log(what);
		if (Statics.verbose > 0)
			System.out.println(what);
	}

	public void progressCallbackNNodes(long nNodes) throws IOException {
		String what = "Total nodes searched: " + nNodes;

		Statics.log(what);
		if (Statics.verbose > 0)
			System.out.println(what);
	}

	public void progressCallbackDepthReached(int depth) throws IOException {
		String what = "Depth reached: " + depth + " (thread: " + Thread.currentThread().getName() + ")";
		Statics.log(what);
		if (Statics.verbose > 0)
			System.out.println(what);
	}

	public void progressCallbackCurrMove(Move move) throws IOException {
		Statics.log("Current move: " + move);
	}

	ReturnData getMove(Scene scene, PlayerColor currentPlayer) throws IOException {
		Move move;
		int c = 0;

		do {
			BufferedReader in = new BufferedReader(new InputStreamReader(System.in));

			do {
				StringBuffer error = new StringBuffer("");
				String moveString;

				System.out.print("Your move: ");

				do {
					String line = in.readLine();
					if (line == null)
						return null;

					moveString = line.trim().toLowerCase();

					if (moveString.equals("quit"))
						return new ReturnData(ReturnType.RET_QUIT, null);
					if (moveString.equals("undo"))
						return new ReturnData(ReturnType.RET_UNDO, null);
				}
				while(moveString.length() < 2);

				move = Statics.SANToMove(moveString, scene.getBoard(), currentPlayer, scene.getMoveList(currentPlayer), error);
				if (move == null) {
					move = Statics.LANToMove(moveString, scene.getBoard(), currentPlayer, scene.getMoveList(currentPlayer), error);
				}
				if (move == null) {
					System.out.println("Error parsing move: " + error);
				}
			}
			while(move == null);

			System.out.print("Move " + Statics.moveToString(move, false));
			System.out.println("? (y/n)");

			do {
				c = Character.toUpperCase((char)in.read());
			}
			while(c != 'Y' && c != 'N');
		}
		while(c != 'Y');

		return new ReturnData(ReturnType.RET_MOVE, move);
	}

	public IO_Console(String pgnFile, int maxDepth, int maxExtension, boolean random, double maxDuration, String boardInit, final Brain brain, final boolean ponder, List<String> beepers) throws Exception {
		super(maxDepth, maxExtension, random, maxDuration, brain, ponder, false, beepers);

		if (boardInit != null) {
			Object [] epd = Statics.EPDToScene(boardInit);
			scene = (Scene)epd[0];
			currentPlayer = (PlayerColor)epd[1];
		}
		else {
			scene = new Scene();
		}

		this.pgnFile = pgnFile;
	}

	public void loop() throws IOException, InterruptedException {
		int it = 0;
		double previousShannonValue = 0;
		PlayerColor humanColor = PlayerColor.WHITE;
		Pgn pgnObject = null;
		int c;

		Statics.log(Statics.startupLine);

		if (bp != null)
			bp.startPonder(1, scene, maxDepth, currentPlayer, moveRepHist, 0, 0, null);

		System.out.println("Play with black or white? (b/w)");
		do {
			c = Character.toUpperCase(System.in.read());
		}
		while(c != 'W' && c != 'B');

		if (c == 'B') {
			humanColor = PlayerColor.BLACK;
			Statics.log("Opponent plays with black");
			pgnObject = new Pgn(PlayerColor.WHITE, pgnFile);
		}
		else {
			Statics.log("Opponent plays with white");
			pgnObject = new Pgn(PlayerColor.BLACK, pgnFile);
		}

		Move lastMove = null;

	undo_loop:
		for(;;) {
			if (currentPlayer == PlayerColor.WHITE) {
				it++;

				System.out.println("Move #" + it);
			}

			scene.validateMoves(currentPlayer);
			// System.out.println("HIER");
			// for(Move cur : scene.getMoveList(currentPlayer)) {
			// 	System.out.println("" + cur);
			// }
			// System.out.println("daar");

			if (scene.isCheckMate(PlayerColor.WHITE)) {
				System.out.println("White is check-mate");
				Statics.log("White is check-mate");
				pgnObject.closePgnFile("0-1");
				break;
			}
			if (scene.isCheckMate(PlayerColor.BLACK)) {
				System.out.println("Black is check-mate");
				Statics.log("Black is check-mate");
				pgnObject.closePgnFile("1-0");
				break;
			}

			//Statics.displayBoard(scene.getBoard());
			String fenStr = Statics.toFENString(scene.getBoard(), currentPlayer, null, it, false);
			System.out.println("FEN: " + fenStr);

			MoveResult move = null;

			if (currentPlayer == humanColor) {
				ReturnData ret = getMove(scene, currentPlayer);

				if (ret.getReturnType() == ReturnType.RET_QUIT)
					break;

				if (ret.getReturnType() == ReturnType.RET_UNDO && h.size() >= 1) {
                                        HistoryElement curHe = h.remove(h.size() - 1);

					if (currentPlayer == PlayerColor.WHITE) {
						Object [] epd = Statics.EPDToScene(curHe.getWhiteBeforeFen());
						scene = (Scene)epd[0];
					}
					else {
						Object [] epd = Statics.EPDToScene(curHe.getBlackBeforeFen());
						scene = (Scene)epd[0];
					}

					continue undo_loop;
				}

				lastMove = ret.getMove();

				move = new MoveResult(0, 0, lastMove, "opponent");

				Statics.log("" + it + ". opponent moves: " + Statics.moveToString(lastMove, false));
			}
			else {
				if (bp != null)
					bp.stopPonder();

				move = brain.calculateMove(it, scene, maxDepth, currentPlayer, moveRepHist, (long)(maxDuration * 1000.0), maxExtension, lastMove, false);

				String out = Statics.moveToString(move.getMove(), false) + " (" + move.getEval() + ")";

				System.out.print("Will move: " + out);
				System.out.println("");

				Statics.log("" + it + ". POS moves: " + out);
			}

			if (Statics.getVerbosity() > 0) {
				String moveStats = move.toString();

				Statics.log(moveStats);
				System.out.println(moveStats);
			}

			pgnObject.pgnEmitMove(scene.getBoard(), move.getMove(), currentPlayer);

			moveRepHist.push(currentPlayer, brain.getHash(scene.getBoard(), currentPlayer));

			scene = scene.Move(move.getMove());

			System.out.print("Computer: ");

			double shannonValue = move.getEval();

			if (shannonValue - previousShannonValue < 0.0d)
				System.out.print(":-(");
			else if (shannonValue - previousShannonValue > 0.0d)
				System.out.print(":-)");
			else
				System.out.print(":-|");

			System.out.println(" " + shannonValue);

			previousShannonValue = shannonValue;

			currentPlayer = Statics.opponentColor(currentPlayer);

			if (currentPlayer == humanColor && bp != null)
				bp.startPonder(it, scene, maxDepth, currentPlayer, moveRepHist, 0, 0, move.getMove());
		}

		if (bp != null)
			bp.stopPonder();
	}
}
