package com.vanheusden.pfa;

import java.io.*;
import java.util.*;
import java.util.concurrent.atomic.*;

class IO_Testcases extends IO {
	private MoveRepetition moveRepHist = new MoveRepetition();
	int cur = -1;
	int threadsPerSlot = 8;
	int maxNProcs = 4;

	public void progressCallback(String what) {
		Statics.log(what);
	}

	public void progressCallbackDebug(String what) {
		Statics.log(what);
	}

	public void progressCallbackXboardShowThinking(int id, int score, long durationCentiSeconds, long nodesVisited, String details) throws IOException {
		Statics.log("" + id + " " + score + " " + durationCentiSeconds + " " + nodesVisited + " " + details);
	}

	/* public stuff */
	public IO_Testcases(int maxDepth, int maxExtension, boolean random, double maxDuration, final Brain brain, final boolean ponder, List<String> beepers) throws Exception {
		super(maxDepth, maxExtension, random, maxDuration, brain, ponder, false, beepers);

		this.maxDuration = maxDuration;
	}

	public void fail(String what) {
		Statics.log("TEST FAIL: " + what);
		System.err.println(what);

		//System.exit(1);
	}

	void doPerft(final String epd, final long [] cnts) {
		for(int index=0; index<cnts.length; index++) {
			final long cnt = cnts[index];

			if (cnt == 0)
				continue;

			final Perft p = new Perft(false);
			final long result;

			if ((result = p.perft(epd, index + 1, maxNProcs)) != cnt)
				fail("perft: " + result + " (expected: " + cnt + ")");
		}
	}

	void testTPT() throws IOException {
		TranspositionTable tt = new TranspositionTable("/tmp/dbptestcases-tt.dat", 1024 * 1024, this);

		// first put
		int score = 1234;
		byte depth = 1;
		int ply = 0;
		Flag flag = Flag.EXACT;
		Move move = new Move(1, 2, 3, 4);
		for(int hash=0; hash<1024; hash++)
			tt.put(hash, move, score, depth, flag, ply);

		for(int hash=0; hash<1024; hash++) {
			TTElement objs = tt.get(hash);

			short scoreThen = objs.getEval(0);
			short depthThen = objs.getDepth();

			if (scoreThen != score)
				fail("(a) invalid score " + scoreThen + ", expected " + score);

			if (depthThen != depth)
				fail("(a) invalid depth " + depthThen + ", expected " + depth);
		}

		// second
		int hash = 9999;
		ply = 9;
		int scoreBase = 5678;
		score = scoreBase - ply; 
		tt.put(hash, move, score, depth, flag, ply);
		// get
		TTElement objs = tt.get(hash);
		short scoreThen = objs.getEval(0);
		if (scoreThen != scoreBase)
			fail("(b) invalid score " + scoreThen + ", expected: " + scoreBase);
	}

        void testQuiesce() throws InterruptedException, IOException {
                Object [] epd = Statics.EPDToScene("rnbqkbnr/p1p1p1p1/3p4/1pP1Pp2/6P1/8/1P3P1P/RNBQKBNR w KQkq - 0 1");
                scene = (Scene)epd[0];
                currentPlayer = (PlayerColor)epd[1];

                AtomicBoolean flag = new AtomicBoolean(false);
                AlphaBeta abIn = new AlphaBeta();
                AlphaBeta res = brain.quiesceSearch(scene, 0, currentPlayer, abIn, flag, 0);
        }

	long start(int nr) {
		cur = nr;

		Statics.log("Test " + nr + " start");
		System.out.print("" + nr + ". ");

		return System.currentTimeMillis();
	}

	double end(long ts) {
		double took = (System.currentTimeMillis() - ts) / 1000.0;

		String ok = " OK (" + took + "s)";

		System.out.println(ok);
		Statics.log("Test " + cur + ": " + ok);

		return took;
	}

	public void loop() throws IOException, InterruptedException {
		try {
			double tt = -1.0;
			long t = -1;

if (false) {
			t = start(1);
			doPerft(null, new long [] { 20, 400, 8902, 197281 });
			tt += end(t);

			// http://www.rocechess.ch/perft.html
			t = start(2);
			doPerft("n1n5/PPPk4/8/8/8/8/4Kppp/5N1N b - -", new long [] { 24, 496, 9483, 182838 });
			tt += end(t);
			// ^ //

			t = start(3);
			doPerft("8/3K4/2p5/p2b2r1/5k2/8/8/1q6 b - -", new long [] { 50, 279 });
			tt += end(t);

			t = start(4);
			doPerft("r1b4r/p2p1ppp/1ppbn2k/4P2P/1PB5/2P2PP1/P7/RNKQ3R w - -", new long [] { 0, 0, 0, 1066400 });
			tt += end(t);

			t = start(5);
			doPerft("1q5k/8/1p2n3/p1p1rb2/2Pp3n/PP1P2P1/3B1R2/2R2QK1 b - -", new long [] { 0, 0, 0, 1233636 });
			tt += end(t);

			t = start(6);
			doPerft("r3kb1r/pp1bpppp/8/5n2/4n3/4B3/PPP2PPP/RN1K1B1R w kq -", new long [] { 0, 0, 0, 1240989 });
			tt += end(t);

			// http://chessprogramming.wikispaces.com/Perft+Results
			t = start(7);
			doPerft("r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq -", new long [] { 48, 2039, 97862 });
			tt += end(t);

			t = start(8);
			doPerft("8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - -", new long [] { 14, 191, 2812, 43238 });
			tt += end(t);

			t = start(9);
			doPerft("r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq -", new long [] { 6, 264, 9467, 422333 });
			tt += end(t);

			t = start(10);
			doPerft("rnbqkb1r/pp1p1ppp/2p5/4P3/2B5/8/PPP1NnPP/RNBQK2R w KQkq -", new long [] { 42, 1352, 53392 });
			tt += end(t);

			t = start(11);
			doPerft("r4rk1/1pp1qppp/p1np1n2/2b1p1B1/2B1P1b1/P1NP1N2/1PP1QPPP/R4RK1 w - -", new long [] { 46, 2079, 89890, 3894594 });
			tt += end(t);
			// ^ //

			// http://www.stmintz.com/ccc/index.php?id=489328
			t = start(12);
			doPerft("7k/8/8/8/1p6/8/PP6/7K w - -", new long [] { 6, 25, 176, 1081, 8479, 57228 });
			tt += end(t);
			// ^ //

			// http://www.chessprogramming.net/philosophy/perfect-perft/
			// Illegal ep move #1
			t = start(15);
			doPerft("3k4/3p4/8/K1P4r/8/8/8/8 b - - 0 1", new long [] { 0, 0, 0, 0, 0, 1134888 });
			tt += end(t);
			// Illegal ep move #2
			t = start(16);
			doPerft("8/8/4k3/8/2p5/8/B2P2K1/8 w - - 0 1", new long [] { 0, 0, 0, 0, 0, 1015133 });
			tt += end(t);

			// missing: EP capture checks opponent
			t = start(17);
			doPerft("8/8/1k6/2b5/2pP4/8/5K2/8 b - d3 0 1", new long [] { 0, 0, 0, 0, 0, 1440467 });
			tt += end(t);

			// short castling gives check
			t = start(18);
			doPerft("5k2/8/8/8/8/8/8/4K2R w K - 0 1", new long [] { 0, 0, 0, 0, 0, 661072 });
			tt += end(t);

			// Long castling gives check
			t = start(19);
			doPerft("3k4/8/8/8/8/8/8/R3K3 w Q - 0 1", new long [] { 0, 0, 0, 0, 0, 803711 });
			tt += end(t);

			// Castle rights
			t = start(20);
			doPerft("r3k2r/1b4bq/8/8/8/8/7B/R3K2R w KQkq - 0 1", new long [] { 0, 0, 0, 1274206 });
			tt += end(t);

			// Castling prevented
			t = start(21);
			doPerft("r3k2r/8/3Q4/8/8/5q2/8/R3K2R b KQkq - 0 1", new long [] { 0, 0, 0, 1720476 });
			tt += end(t);

			// Promote to give check
			t = start(22);
			doPerft("4k3/1P6/8/8/8/8/K7/8 w - - 0 1", new long [] { 0, 0, 0, 0, 0, 217342 });
			tt += end(t);

			t = start(23);
			doPerft("8/P1k5/K7/8/8/8/8/8 w - - 0 1", new long [] { 0, 0, 0, 0, 0, 92683 });
			tt += end(t);
			// ^ //

			t = start(30);
			testTPT();
			tt += end(t);
}

			t = start(31);
			testQuiesce();
			tt += end(t);

			Statics.log("Running testcases took in total " + tt + " seconds");
		}
		catch(Exception e) {
			Statics.logException("Testcases", e);

			fail("unexpected exception: " + e);
		}
	}
}
