package com.vanheusden.pfa;

import java.io.*;
import java.net.*;
import java.util.*;

class IO_UCI_socket extends IO_UCI {
	ServerSocket ss = null;
	Socket client = null;
	
	public void progressCallback(String what) {
		if (Statics.verbose > 0)
			emit("info string " + what);
	}

	public void progressCallbackDebug(String what) {
		emit("info string " + what);
	}

	IO_UCI_socket(String pgnFile, int maxDepth, int maxExtension, boolean random, double maxDuration, final int port, final Brain brain, final boolean ponder, final boolean ignore_protocol_overrides, List<String> beepers) throws Exception {
		super(pgnFile, maxDepth, maxExtension, random, maxDuration, brain, ponder, ignore_protocol_overrides, beepers);

		scene = new Scene();

		ss = new ServerSocket(port);
		ss.setReuseAddress(true);
	}
	
	void close() {
		try {
			if (xboardPipe != null)
				xboardPipe.close();

			if (output != null)
				output.close();

			if (client != null)
				client.close();
		}
		catch(IOException ie) {
			Statics.log("IO_UCI_socket: close failed " + ie);
		}
		
		xboardPipe = null;
		output = null;
		client = null;
	}

	void setupIO() throws IOException {
	}

	public int getPort() {
		return ss.getLocalPort();
	}

	void reconnect() {
		Statics.log("IO_UCI_socket::reconnect()");

		for (; ss != null;) {
			try {
				client = ss.accept();
				Statics.log("Connected with " + client.getRemoteSocketAddress());

				client.setTcpNoDelay(true);

				xboardPipe = new BufferedReader(new InputStreamReader(client.getInputStream()));

				output = new BufferedWriter(new OutputStreamWriter(client.getOutputStream()));

				break;
			} catch (IOException ie) {
				Statics.log("IO_UCI_socket::reconnect " + ie);
			}
		}
	}

	synchronized void emit(String what) {
		if (output == null) // should not occur
			reconnect();

		if (output != null) {
			try {
				// no newLine because then the 'what' could succeed with the
				// newLine() failing causing a duplicate 'what' after reconnect
				output.write(what + "\n");
				output.flush();

				Statics.log("socket out: " + what);

			} catch (IOException ie) {
				Statics.log("IO_UCI_socket::emit " + ie);

				close();
			}
		}
	}

	String get() {
		for(;;) {
			if (xboardPipe == null)
				reconnect();

			try {
				String in = xboardPipe.readLine();

				Statics.log("socket in: " + in);

				return in;
			}
			catch (IOException ie) {
				Statics.log("IO_UCI_socket::get " + ie);

				close();
			}
		}
	}
}
