package com.vanheusden.pfa;

import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;
import java.util.concurrent.locks.*;

public class MySem extends Semaphore {
	long totalGetWait, cntGetWait;
	long startHeld;
	long totalHeld, cntHeld;
	String name;
	AtomicInteger nWaiting = new AtomicInteger(0);
	long nWaitingTotal;
	AtomicLong nTryAcq = new AtomicLong(0);
	long nTryAcqSucceeded;

	public MySem(final int depth, final boolean fair, final String name) {
		super(depth, fair);

		totalGetWait = cntGetWait = 0;
		totalHeld = cntHeld = 0;

		nWaitingTotal = nTryAcqSucceeded = 0;

		this.name = name;
	}

	public void destructor() {
		System.err.println("semaphore " + name + ", acq duration: " + totalGetWait / (double)cntGetWait + "ns, held duration: " + totalHeld / (double)cntHeld + "ns, n waiting: " + nWaitingTotal / (double)cntGetWait + ", try acq: " + nTryAcq.get() + ", try succeeded: " + nTryAcqSucceeded);
	}

	public void acquireUninterruptibly() {
		long now1 = System.nanoTime();

		nWaiting.incrementAndGet();

		super.acquireUninterruptibly();

		long now2 = System.nanoTime();

		nWaiting.decrementAndGet();
		nWaitingTotal += nWaiting.get();

		startHeld = now2;

		totalGetWait += now2 - now1;
		cntGetWait++;
	}

	public void release() {
		long now = System.nanoTime();

		totalHeld += now - startHeld;
		cntHeld++;

		super.release();
	}

	public boolean tryAcquire() {
		nTryAcq.incrementAndGet();

		boolean succeeded = super.tryAcquire();

		if (succeeded) {
			startHeld = System.nanoTime();

			nTryAcqSucceeded++;
		}

		return succeeded;
	}
}
